<?php

namespace Ranig;

use ManiaControl\Callbacks\CallbackListener;
use ManiaControl\Callbacks\CallbackManager;
use ManiaControl\Callbacks\Callbacks;
use ManiaControl\Callbacks\Structures\TrackMania\OnWayPointEventStructure;
use ManiaControl\Callbacks\Structures\TrackMania\OnStuntEventStructure;
use ManiaControl\Callbacks\Structures\TrackMania\OnRespawnStructure;
use ManiaControl\Callbacks\Structures\Common\BasePlayerTimeStructure;
use ManiaControl\Callbacks\TimerListener;
use ManiaControl\Commands\CommandListener;
use ManiaControl\Logger;
use ManiaControl\ManiaControl;
use ManiaControl\Players\Player;
use ManiaControl\Players\PlayerActions;
use ManiaControl\Players\PlayerManager;
use ManiaControl\Plugins\Plugin;
use ManiaControl\Statistics\StatisticCollector;
use ManiaControl\Statistics\StatisticManager;
use ManiaControl\Utils\Formatter;
use Maniaplanet\DedicatedServer\Structures\AbstractStructure;

/**
 */
class KickIdlePlayersPlugin implements Plugin, CallbackListener, CommandListener, TimerListener {
	/*
	 * Constants
	 */
	const PLUGIN_ID              = 137;
	const PLUGIN_VERSION         = 1;
	const PLUGIN_NAME            = 'Kick Idle Players Plugin';
	const PLUGIN_AUTHOR          = 'Ranig';
	const SETTING_IDLE_COUNT     = 'Number of rounds before a player is kicked';
	const SETTING_FORCE_SPEC     = 'Force spectator instead of kick';
	const SETTING_USE_TIME_LIMIT = 'Sets up a time limit for idle players';
	const SETTING_TIME_LIMIT     = 'Time limit after which an idle player is kicked';
	const SETTING_ACTIVE_STUNT   = 'Count stunts to evaluate if a player is idle';

	/*
	 * Private properties
	 */
	/** @var ManiaControl $maniaControl * */
	private $maniaControl = null;
	/** @var Array $idlePlayerList * */
	private $idlePlayerList = null;
	private $activePlayerList = null;
	private $giveUpPlayerList = null;
	private $idleLimit = null;

	/**
	 * @see \ManiaControl\Plugins\Plugin::prepare()
	 */
	public static function prepare(ManiaControl $maniaControl) {
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getId()
	 */
	public static function getId() {
		return self::PLUGIN_ID;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getName()
	 */
	public static function getName() {
		return self::PLUGIN_NAME;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getVersion()
	 */
	public static function getVersion() {
		return self::PLUGIN_VERSION;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getAuthor()
	 */
	public static function getAuthor() {
		return self::PLUGIN_AUTHOR;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getDescription()
	 */
	public static function getDescription() {
		return "KickIdlePlayers Plugin, kicks idle players after a number of maps or a predefined time";
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::load()
	 */
	public function load(ManiaControl $maniaControl) {
		$this->maniaControl = $maniaControl;

		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_IDLE_COUNT, 2);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_FORCE_SPEC, false);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_USE_TIME_LIMIT, false);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TIME_LIMIT, 300);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_ACTIVE_STUNT, false);

		// Callbacks
		$this->maniaControl->getCallbackManager()->registerCallbackListener(PlayerManager::CB_PLAYERCONNECT, $this, 'handlePlayerConnect');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(PlayerManager::CB_PLAYERDISCONNECT, $this, 'handlePlayerDisconnect');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::BEGINMAP, $this, 'handleBeginMap');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::ENDMAP, $this, 'handleEndMap');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_ONWAYPOINT, $this, 'handlePlayerCheckpoint');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_ONSTUNT, $this, 'handlePlayerStunt');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_ONRESPAWN, $this, 'handlePlayerRespawn');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_ONGIVEUP, $this, 'handlePlayerGiveUp');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_ONFINISHLINE, $this, 'handlePlayerFinish');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(Callbacks::TM_ONSTARTCOUNTDOWN, $this, 'handleStart');

		// TODO: array
		$this->idlePlayerList = array();
		$this->activePlayerList = array();
		$this->giveUpPlayerList = array();
		$this->idleLimit = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_IDLE_COUNT);
		if ($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_USE_TIME_LIMIT)) {
			$this->idleLimit = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_TIME_LIMIT) / 15;
		}

		foreach ($this->maniaControl->getPlayerManager()->getPlayers() as $player) {
			$this->idlePlayerList[$player->login] = 0;
			$this->giveUpPlayerList[$player->login] = false;
			$this->activePlayerList[$player->login] = 1;
		}

		$this->maniaControl->getTimerManager()->registerTimerListening($this, 'handleTimeIncrement', 15000);
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::unload()
	 */
	public function unload() {
		$this->idlePlayerList = array();
		$this->activePlayerList = array();
		$this->giveUpPlayerList = array();
		$this->maniaControl->getTimerManager()->unregisterTimerListening($this, 'handleTimeIncrement');
	}

	/**
	 * Sets a player as active
	 *
	 * @param $login the login of the player to activate
	 */
	public function activatePlayer($login) {
		$this->activePlayerList[$login] = $this->activePlayerList[$login] + 1;
		$this->giveUpPlayerList[$login] = false;
	}

	/**
	 * Handle PlayerConnect callback
	 *
	 * @param Player $player
	 */
	public function handlePlayerConnect(Player $player) {
		$this->idlePlayerList[$player->login] = 0;
		$this->giveUpPlayerList[$player->login] = false;
		$this->activePlayerList[$player->login] = 1;
	}

	/**
	 * Handle PlayerDisconnect callback
	 *
	 * @param Player $player
	 */
	public function handlePlayerDisconnect(Player $player) {
		unset($this->idlePlayerList[$player->login]);
		unset($this->activePlayerList[$player->login]);
		unset($this->giveUpPlayerList[$player->login]);
	}

	/**
	 * Passing through a checkpoint flags a player as active
	 */ 
	public function handlePlayerCheckpoint(OnWayPointEventStructure $structure) {
		$playerLogin = $structure->getLogin();
		$this->activatePlayer($playerLogin);
	}

	/**
	 * Performing a stung flags a player as active
	 */ 
	public function handlePlayerStunt(OnStuntEventStructure $structure) {
		if ($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_ACTIVE_STUNT)) {
			$playerLogin = $structure->getLogin();
			$this->activatePlayer($playerLogin);
		}
	}

	/**
	 * Respawining flags a player as active
	 */ 
	public function handlePlayerRespawn(OnRespawnStructure $structure) {
		$playerLogin = $structure->getLogin();
		$this->activatePlayer($playerLogin);
	}

	/**
	 * Giving up places the player in the giveUpPlayerList so that it doesn't get kicked
	 */ 
	public function handlePlayerGiveUp(BasePlayerTimeStructure $structure) {
		$playerLogin = $structure->getLogin();
		$this->giveUpPlayerList[$playerLogin] = true;
	}

	/**
	 * Finishing a map flags the player as active
	 */ 
	public function handlePlayerFinish(OnWayPointEventStructure $structure) {
		$playerLogin = $structure->getLogin();
		$this->activatePlayer($playerLogin);
	}

	/**
	 * Removes players from the giveUpPlayerList at the start
	 */ 
	public function handleStart() {
		foreach ($this->maniaControl->getPlayerManager()->getPlayers() as $player) {
			$this->giveUpPlayerList[$player->login] = false;
		}
	}

	/**
	 * Reset the activity counters for a Player and manages special cases
	 */
	public function checkIdlePlayer(Player $player) {
		$this->activePlayerList[$player->login] = 0;
		
		// do not kick spectators
		if ($player->isSpectator) {
			$this->idlePlayerList[$player->login] = 0;
		}

		// do not kick players who give up
		if ($this->giveUpPlayerList[$player->login]) {
			$this->idlePlayerList[$player->login] = 0;
		}

		// warn if the player reaches its idle limit
		if ($this->idlePlayerList[$player->login] == $this->idleLimit - 1) {
			$message = '$f00Warning! $fffYou will be kicked if you remain idle!';
			$this->maniaControl->getChat()->sendChat($message, $player);
		}
	}
	
	/**
	 * Increments the idle counter for a Player if it hasn't been active on the last time period or map
	 */ 
	public function incrementIdlePlayer(Player $player) {
		if ($this->activePlayerList[$player->login] == 0) {
			$this->idlePlayerList[$player->login] = $this->idlePlayerList[$player->login] + 1;
		} else {
			$this->idlePlayerList[$player->login] = 0;
		}
	}

	/**
	 * Kicks or force spec the Player if it has reached its idle limit
	 */ 
	public function kickIdlePlayer(Player $player) {
		if ($this->idleLimit < 2) {
			return;
		}
		
		if ($this->activePlayerList[$player->login] == 0) {
			if ($this->idlePlayerList[$player->login] >= $this->idleLimit) {
				if ($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_FORCE_SPEC)) { 
					$spec = $this->maniaControl->getPlayerManager()->getPlayerActions()->forcePlayerToSpectator('KickIdlePlayersPlugin', $player->login, PlayerActions::SPECTATOR_BUT_KEEP_SELECTABLE, true, false);
					$this->idlePlayerList[$player->login] = 0;
					$this->activePlayerList[$player->login] = 1;
				} else {
					$kick = $this->maniaControl->getPlayerManager()->getPlayerActions()->kickPlayer('KickIdlePlayersPlugin', $player->login, 'You have been warned!', false);
				}
			}
		}
	}

	/**
	 * Increments idle count on begin map
	 */
	public function handleBeginMap() {
		$this->idleLimit = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_IDLE_COUNT);
		if ($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_USE_TIME_LIMIT)) {
			$this->idleLimit = $this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_TIME_LIMIT) / 15;
			return;
		}
		
		foreach ($this->maniaControl->getPlayerManager()->getPlayers() as $player) {
			$this->checkIdlePlayer($player);
		}		
	}
	
	/**
	 * Checks activity and kicks players on each time increment (if active)
	 */
	public function handleTimeIncrement() {
		if (!$this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_USE_TIME_LIMIT)) {
			return;
		}

		foreach ($this->maniaControl->getPlayerManager()->getPlayers() as $player) {
			$this->kickIdlePlayer($player);
			$this->incrementIdlePlayer($player);
			$this->checkIdlePlayer($player);
		}
	}

	/**
	 * Kick idle players on end map
	 */
	public function handleEndMap() {
		if ($this->maniaControl->getSettingManager()->getSettingValue($this, self::SETTING_USE_TIME_LIMIT)) {
			return;
		}

		foreach ($this->maniaControl->getPlayerManager()->getPlayers() as $player) {
			$this->incrementIdlePlayer($player);
			$this->kickIdlePlayer($player);
		}
	}

}